"""
Pig allele names from:
    Analyzing the genetic characteristics and function of the swine
    leukocyte antigen 2 gene in a Chinese inbreed of pigs
"""

from mhcgnomes import (
    Allele,
    parse,
)

from .common import eq_


def test_SLA_1_0101_with_seps():
    eq_(parse("SLA-1*01:01"), Allele.get("SLA", "1", "01", "01"))


def test_SLA_1_HB01():
    eq_(parse("SLA-1-HB01"), Allele.get("SLA", "1", "HB01"))


def test_SLA_1_0101_no_seps():
    eq_(parse("SLA-10101"), Allele.get("SLA", "1", "01", "01"))


def test_SLA_2_07we01_no_normalization():
    # SLA-2*07we01 is the provisional allele name for SLA-2*07:03
    eq_(parse("SLA-2*07we01", use_allele_aliases=False), Allele.get("SLA", "2", "07we01"))


def test_SLA_2_07we01_normalize():
    # SLA-2*07we01 is the provisional allele name for SLA-2*07:03
    eq_(parse("SLA-2*07we01", use_allele_aliases=True), Allele.get("SLA", "2", "07", "03"))


def test_SLA_2_jh01_no_normalization():
    # SLA-2*jh01 is the provisional allele name for 2*15:01
    eq_(parse("SLA-2*jh01", use_allele_aliases=False), Allele.get("SLA", "2", "jh01"))


def test_SLA_2_jh01_normalize():
    # SLA-2*jh01 is the provisional allele name for 2*15:01
    eq_(parse("SLA-2*jh01", use_allele_aliases=True), Allele.get("SLA", "2", "15", "01"))


def test_SLA_2_w09pt22_no_normalization():
    # SLA-2*w09pt22 is the provisional allele name for 2*09:03
    eq_(parse("SLA-2*w09pt22", use_allele_aliases=False), Allele.get("SLA", "2", "w09pt22"))


def test_SLA_2_w09pt22_normalize():
    # SLA-2*w09pt22 is the provisional allele name for 2*09:03
    eq_(parse("SLA-2*w09pt22", use_allele_aliases=True), Allele.get("SLA", "2", "09", "03"))


def test_parse_to_string_SLA_1_CHANGDA():
    eq_(parse("SLA-1-CHANGDA", use_allele_aliases=False).to_string(), "SLA-1*CHANGDA")


def test_parse_to_string_SLA_1_HB01():
    eq_(parse("SLA-1-HB01", use_allele_aliases=False).to_string(), "SLA-1*HB01")


def test_parse_to_string_SLA_1_HB01_with_colon_sep():
    eq_(parse("SLA-1:HB01", use_allele_aliases=False).to_string(), "SLA-1*HB01")


def test_parse_to_string_SLA_1_TPK():
    eq_(parse("SLA-1-TPK", use_allele_aliases=False).to_string(), "SLA-1*TPK")


def test_parse_to_string_SLA_2_TPK_AA():
    eq_(parse("SLA-2-TPK.AA", use_allele_aliases=True).to_string(), "SLA-2*TPK")


def test_parse_to_string_SLA_2_TPK_AA_with_colon_sep():
    eq_(parse("SLA-2:TPK.AA", use_allele_aliases=True).to_string(), "SLA-2*TPK")


def test_parse_to_string_SLA_2_LWH():
    eq_(parse("SLA-2-LWH", use_allele_aliases=False).to_string(), "SLA-2*LWH")


def test_parse_to_string_SLA_2_YDY01():
    eq_(parse("SLA-3-YDY01", use_allele_aliases=False).to_string(), "SLA-3*YDY01")



from mhcgnomes import NormalizingDictionary

from .common import eq_


def test_test_NormalizingDictionary_from_dict_to_dict_identity():
    d = {"aA--": 1, "bb": 2}
    eq_(d, NormalizingDictionary.from_dict(d).to_dict())


def test_NormalizingDictionary_case_invariant():
    d = NormalizingDictionary.from_dict({"aA--": 1, "bb": 2})
    eq_(d["AA--"], 1)
    eq_(d["BB"], 2)


def test_NormalizingDictionary_dash_invariant():
    d = NormalizingDictionary.from_dict({"aA--": 1, "bb": 2})
    eq_(d["AA"], 1)
    eq_(d["BB-"], 2)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .errors import ParseError

class1_subtypes = {"Ia", "Ib", "Ic", "Id"}

class2_subtypes = {
    "IIa",
    "IIb",
}

class1_restrictions = {"I"}.union(class1_subtypes)
class2_restrictions = {"II"}.union(class2_subtypes)

valid_class_restrictions = class1_restrictions.union(class2_restrictions)

classical_subtypes = {"Ia", "IIa"}


def mhc_class_is_more_specific(original_class=None, new_class=None):
    return (
        (original_class is None and new_class is not None)
        or (original_class == "I" and new_class in class1_subtypes)
        or (original_class == "II" and new_class in class2_subtypes)
    )


def is_class1(mhc_class):
    return mhc_class in class1_restrictions


def is_class2(mhc_class):
    return mhc_class in class2_restrictions


def is_valid_restriction(original_class=None, new_class=None):
    if original_class is None:
        return True

    if new_class is None:
        # once we've restricted, can't go backwards
        return False

    if original_class == new_class:
        return True

    return mhc_class_is_more_specific(original_class, new_class)


def restrict_alleles(alleles, mhc_class):
    if mhc_class == "I":
        valid_subtypes = class1_restrictions
    elif mhc_class == "II":
        valid_subtypes = class2_restrictions
    else:
        valid_subtypes = {mhc_class}

    return [allele for allele in alleles if allele.mhc_class in valid_subtypes]


def normalize_mhc_class_string(mhc_class, raise_on_error=True):
    original_string = mhc_class
    mhc_class = mhc_class.lower()
    if mhc_class.startswith("class-"):
        _, mhc_class = mhc_class.split("class-")
    elif mhc_class.startswith("class "):
        _, mhc_class = mhc_class.split("class ")
    mhc_class = mhc_class.replace("i", "I")
    mhc_class = mhc_class.replace("1", "I")
    mhc_class = mhc_class.replace("2", "II")
    if mhc_class not in valid_class_restrictions:
        if raise_on_error:
            raise ParseError(f"Invalid MHC class: '{original_string}'")
        else:
            return None
    return mhc_class



from mhcgnomes import Allele
from mhcgnomes.mhc_class_helpers import is_valid_restriction, restrict_alleles

from .common import eq_


def test_is_valid_restriction_class2_subtypes():
    assert is_valid_restriction("II", "IIa")
    assert is_valid_restriction("II", "IIb")
    assert not is_valid_restriction("II", "Ia")
    assert not is_valid_restriction("IIa", "II")


def test_restrict_alleles_class2_includes_ii_iia_iib():
    alleles = [
        Allele.get("HLA", "DRA", "01", "01"),
        Allele.get("HLA", "DOA", "01", "01"),
        Allele.get("HLA", "DRB6", "01", "01"),
    ]
    restricted = restrict_alleles(alleles, "II")
    eq_(set(restricted), set(alleles))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from serializable import Serializable

from .token_canonical_sequences import (
    ALPHA_CHAIN_TOKEN_SEQ,
    BETA_CHAIN_TOKEN_SEQ,
    CLASS1_TOKEN_SEQ,
    CLASS2_TOKEN_SEQ,
    MUTANT_TOKEN_SEQ,
)


class Token(Serializable):
    def __init__(self, seq, raw_string=None):
        self.seq = seq
        if raw_string is None:
            raw_string = seq
        self.raw_string = raw_string

    def __str__(self):
        return f"Token('{self.seq}')"

    def __repr__(self):
        return f"Token(seq='{self.seq}', raw_string='{self.raw_string}')"

    def __eq__(self, other):
        if type(other) is Token:
            return self.seq == other.seq
        elif type(other) is str:
            return self.seq == other
        else:
            return False

    def __hash__(self):
        return hash(self.seq)

    def __lt__(self, other):
        if type(other) is Token:
            return self.seq < other.seq
        elif type(other) is str:
            return self.seq < other
        else:
            return False

    @property
    def is_class1(self):
        return self.seq == CLASS1_TOKEN_SEQ

    @property
    def is_class2(self):
        return self.seq == CLASS2_TOKEN_SEQ

    @property
    def is_class1_or_class2(self):
        return self.is_class1 or self.is_class2

    @property
    def is_alpha(self):
        return self.seq == ALPHA_CHAIN_TOKEN_SEQ

    @property
    def is_beta(self):
        return self.seq == BETA_CHAIN_TOKEN_SEQ

    @property
    def is_alpha_or_beta(self):
        return self.is_alpha or self.is_beta

    @property
    def is_mutant(self):
        return self.seq == MUTANT_TOKEN_SEQ

    @property
    def is_slash(self):
        return self.seq == "/"

    @property
    def is_haplotype(self):
        return self.seq == "haplotype"

    @property
    def is_gene(self):
        return self.seq == "gene"

    @property
    def is_allele(self):
        return self.seq == "allele"

    @property
    def is_serotype(self):
        return self.seq == "serotype"

    @property
    def is_supertype(self):
        return self.seq == "supertype"

    @property
    def can_be_identifier(self):
        return not (
            self.is_alpha_or_beta
            or self.is_class1_or_class2
            or self.is_slash
            or self.is_haplotype
            or self.is_gene
            or self.is_allele
            or self.is_serotype
            or self.is_supertype
        )



import pytest

from mhcgnomes import (
    MhcClass,
    parse,
)


@pytest.mark.parametrize(
    "s",
    [
        "human class 1",
        "human class i",
        "human class I",
        "hla class 1",
        "hla class i",
        "hla class I",
        "HLA class 1",
        "HLA class i",
        "HLA class I",
    ],
)
def test_human_class_1(s):
    """Test parsing of human class I MHC strings."""
    expected_parsed_result = MhcClass.get("HLA", "I")
    expected_string_repr = "human class I"
    parsed_result = parse(s)
    assert parsed_result == expected_parsed_result, (
        f"Expected {expected_parsed_result} for parsing of '{s}' but got {parsed_result}"
    )
    normalized_str = parsed_result.to_string()
    assert normalized_str == expected_string_repr, (
        f"Expected '{expected_string_repr}' for normalized representation of '{s}' "
        f"but got '{normalized_str}'"
    )
    compact_str = parsed_result.compact_string()
    assert compact_str == expected_string_repr, (
        f"Expected '{expected_string_repr}' for compact representation of '{s}' "
        f"but got '{compact_str}'"
    )



from mhcgnomes import Haplotype, Parser, Species, parse

from .common import eq_


def test_H2_a_in_species_haplotypes_dictionary():
    species = Species.get("H2")
    assert species is not None
    assert "a" in species.haplotypes


def test_get_H2a_haplotype():
    parser = Parser()
    result = parser.get_haplotype("H2", "a")
    assert result is not None
    eq_(result.name, "a")
    assert result.species.is_mouse


def test_parse_H2a_haplotype():
    result = parse("H2a")
    assert result is not None
    eq_(result.name, "a")
    assert result.species.is_mouse


def test_parse_BF19_haplotype():
    result = parse("BF19")
    print(result)
    assert result is not None
    eq_(type(result), Haplotype)
    eq_(result.name, "BF19")
    assert result.species.is_chicken


def test_haplotype_to_string_without_species_has_no_leading_dash():
    result = parse("H2-k", required_result_types=Haplotype)
    assert result is not None
    eq_(result.to_string(include_species=False), "k")



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from os.path import dirname, join

import yaml

from .normalizing_dictionary import NormalizingDictionary

package_dir = dirname(__file__)
data_dir = join(package_dir, "data")


def get_path(yaml_filename):
    return join(data_dir, yaml_filename)


def load(yaml_filename, normalize_first_level_keys=False, normalize_second_level_keys=False):
    path = get_path(yaml_filename)
    with open(path) as f:
        result = yaml.safe_load(f)

    if normalize_second_level_keys:
        # turn first layer of values in dictionary into NormalizingDictionary
        result = {k: NormalizingDictionary.from_dict(v) for (k, v) in result.items()}

    if normalize_first_level_keys:
        result = NormalizingDictionary.from_dict(result)

    return result


# dictionary mapping scientific name -> info dictionary
species = load("species.yaml")

# Dictionary mapping each species prefix to a dictionary from old gene
# names to new ones
gene_aliases = load(
    "gene_aliases.yaml", normalize_first_level_keys=True, normalize_second_level_keys=True
)

# Dictionary mapping each species prefix to a dictionary from
# old/retired/provisional allele names to standard new names
allele_aliases = load(
    "allele_aliases.yaml", normalize_first_level_keys=True, normalize_second_level_keys=True
)

# Dictionary mapping species -> gene -> list of known allele names
known_alleles = load(
    "known_alleles.yaml", normalize_first_level_keys=True, normalize_second_level_keys=True
)

# Dictionary mapping species to haplotype name to list of alleles
haplotypes = load(
    "haplotypes.yaml", normalize_first_level_keys=True, normalize_second_level_keys=True
)

# Dictionary mapping species to serotype name to list of alleles
serotypes = load(
    "serotypes.yaml", normalize_first_level_keys=True, normalize_second_level_keys=True
)

# Dictionary mapping species to heterodimer shorthand to alpha/beta allele pairs
# e.g., HLA -> DQ2.5 -> {alpha: DQA1*05:01, beta: DQB1*02:01}
heterodimers = load(
    "heterodimers.yaml", normalize_first_level_keys=True, normalize_second_level_keys=True
)

# Dictionary mapping species to supertype name to supertype info
# e.g., HLA -> A02 -> {representative: A*02:01, alleles: [A*02:01, A*02:02, ...]}
# Supertypes are functional groupings based on peptide-binding specificity
# (Sidney et al. 2008, https://pmc.ncbi.nlm.nih.gov/articles/PMC2245908/)
supertypes = load(
    "supertypes.yaml", normalize_first_level_keys=True, normalize_second_level_keys=True
)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict

from .common import normalize_string


class NormalizingDictionary:
    """
    Like a regular dictionary but all keys get normalized by a user
    provided function.

    Caution: the number of items in keys() and values() for this dictionary
    may differ because two distinct keys may be transformed to the same
    underlying normalized key and thus will share a value.
    """

    def __init__(self, *pairs, normalize_fn=normalize_string, default_value_fn=None):
        self.store = {}
        self.original_to_normalized_key_dict = {}
        self.normalized_to_original_keys_dict = defaultdict(set)
        self.normalize_fn = normalize_fn
        self.default_value_fn = default_value_fn
        self.update_pairs(pairs)

    def copy(self):
        keys = list(self.keys())
        values = [self.__getitem__(k) for k in keys]
        pairs = zip(keys, values)
        return NormalizingDictionary(
            *pairs, normalize_fn=self.normalize_fn, default_value_fn=self.default_value_fn
        )

    def update_pairs(self, pairs):
        # populate dictionary with initial values via calls to __setitem__
        for k, v in pairs:
            self[k] = v

    def update(self, other_dict):
        self.update_pairs(other_dict.items())

    def __getitem__(self, k):
        k_normalized = self.normalize_fn(k)
        if k_normalized not in self.store:
            if self.default_value_fn is not None:
                self[k] = self.default_value_fn()
            else:
                raise KeyError(k)
        return self.store[k_normalized]

    def __setitem__(self, k, v):
        k_normalized = self.normalize_fn(k)
        self.original_to_normalized_key_dict[k] = k_normalized
        self.normalized_to_original_keys_dict[k_normalized].add(k)
        self.store[k_normalized] = v

    def __contains__(self, k):
        k_normalized = self.normalize_fn(k)
        return k_normalized in self.store

    def original_keys(self, k):
        """
        Returns set of original keys which match the normalized representation
        of the given key.
        """
        k_normalized = self.normalize_fn(k)
        original_keys = self.normalized_to_original_keys_dict.get(k_normalized)
        if original_keys is None:
            return set()
        return original_keys

    def original_key(self, k, default_value=None, pick_best_fn=None):
        """
        Attempts to get a single original key matching which has
        the same the normalized representation as the given input,
        but may raise an exception if there are more than one original
        key that matches.
        """
        ks = list(self.original_keys(k))

        if len(ks) == 0:
            if default_value is None:
                raise KeyError(k)
            else:
                return default_value
        elif len(ks) > 1:
            if pick_best_fn is None:
                raise ValueError(f"Key '{k}' matches multiple entries: {ks}")
            else:
                return pick_best_fn(ks)
        else:
            return ks[0]

    def get(self, k, v=None):
        return self.store.get(self.normalize_fn(k), v)

    def keys(self):
        return self.original_to_normalized_key_dict.keys()

    def normalized_keys(self):
        return self.store.keys()

    def key_sets_aligned_with_values(self):
        """
        Returns all of the original keys associated with each item
        of values
        """
        results = []
        for normalized_key in self.normalized_keys():
            original_keys = self.normalized_to_original_keys_dict[normalized_key]
            results.append((normalized_key, original_keys))
        return results

    def keys_aligned_with_values(self):
        """
        Returns one of the original keys associated with each item
        of values
        """
        for normalized_key, original_keys in self.key_sets_aligned_with_values():
            if len(original_keys) == 0:
                raise ValueError(f"Key set unexpectedly empty for '{normalized_key}'")
            yield sorted(original_keys)[0]

    def values(self):
        return self.store.values()

    def __iter__(self):
        yield from self.keys_aligned_with_values()

    def items(self):
        return zip(self.keys_aligned_with_values(), self.values())

    def to_dict(self):
        return dict(self.items())

    def map_values(self, fn):
        pairs = []
        for k, v in self.items():
            v2 = fn(v)
            pairs.append((k, v2))

        return NormalizingDictionary(
            *pairs, normalize_fn=self.normalize_fn, default_value_fn=self.default_value_fn
        )

    def map_keys(self, fn):
        pairs = [(fn(k), v) for (k, v) in self.items()]
        return NormalizingDictionary(
            *pairs, normalize_fn=self.normalize_fn, default_value_fn=self.default_value_fn
        )

    def invert(self):
        """
        Returns a NormalizingDictionary where every value
        is associated with a set of keys which mapped to it. When
        values are collections (list, set, or tuple) then the elements
        of the collection are turned into individual keys.
        """
        result = NormalizingDictionary(normalize_fn=self.normalize_fn, default_value_fn=set)

        for k, v in self.items():
            if isinstance(v, (list, tuple, set)):
                values = v
            else:
                values = [v]

            for vi in values:
                result[vi].add(k)
        return result

    def __len__(self):
        return len(self.store)

    @classmethod
    def from_dict(cls, d, normalize_fn=normalize_string, default_value_fn=None):
        if type(d) is NormalizingDictionary:
            return d.copy()
        result = cls(*d.items(), normalize_fn=normalize_fn, default_value_fn=default_value_fn)
        if len(d) != 0:
            assert len(result) > 0
            assert len(result.store) > 0
            assert len(result.original_to_normalized_key_dict) > 0
            assert len(result.normalized_to_original_keys_dict) > 0
        return result

    def __str__(self):
        s = (
            f"<NormalizingDictionary with {len(self.store)} unique items, "
            f"normalize_fn={self.normalize_fn}, "
            f"default_value_fn={self.default_value_fn}>"
        )
        all_items = list(self.items())

        for i, (k, v) in enumerate(all_items):
            s += f"\n\t{k}: {v}"
            if i > 10:
                s += "\n..."
                break
        return s

    def __repr__(self):
        return str(self)
